import typing as t
import abc

import pandas as pd


class PydanticModel(t.Protocol):
    @abc.abstractmethod
    def model_validate(self, record: dict) -> t.Self:
        raise NotImplementedError
    
    @abc.abstractmethod
    def model_dump(self) -> dict:
        raise NotImplementedError


class ValidateResult(t.NamedTuple):
    errors: list[Exception]
    errored_df: pd.DataFrame
    success_df: pd.DataFrame


def validate(
    df: pd.DataFrame,
    model: PydanticModel
) -> ValidateResult:
    errors: list[Exception] = []
    errored_df = df.astype('object')
    valid_df = df.astype('object')
    for i, row in df.iterrows():
        record = row.to_dict()
        try:
            valid = model.model_validate(record)
        except Exception as e:
            errors.append(e)
            valid_df.drop(i, inplace=True)
        else:
            valid_df.loc[i] = valid.model_dump()
            errored_df.drop(i, inplace=True)

    errored_df = errored_df.infer_objects()
    valid_df = valid_df.infer_objects()
    return ValidateResult(errors, errored_df, valid_df)


import abraxos
import pandas as pd

from abraxos.extract import ReadCsvResult


def test_read_csv_all() -> None:
    result: ReadCsvResult = abraxos.extract.read_csv_all('tests/bad.csv')
    assert result.bad_lines == [
        ['', '', '', 'd', '', 'f', '', '', '', '', 'f', '', '', '', ''],
        ['', 'f', 'f', '5', '6', '7', '8']
    ]
    assert result.dataframe.equals(
        pd.DataFrame(
            {
                'id': {0: 1, 1: 2, 2: 3, 3: 2, 4: 1},
                'name': {0: 'Odos', 1: 'Kayla', 2: 'Dexter', 3: 'Kayla', 4: 'Odos'},
                'age': {0: '38', 1: '31', 2: 'two', 3: '31', 4: '38'}
            }
        )
    )



import collections.abc as a
import typing as t

import pandas as pd


class ReadCsvResult(t.NamedTuple):
    bad_lines: list[list[str]]
    dataframe: pd.DataFrame
    

def read_csv_chunks(
    path: str,
    chunksize: int,
    **kwargs
) -> a.Generator[ReadCsvResult, None, None]:
    """
    Reads a CSV file in chunks and captures bad lines separately.

    Parameters
    ----------
    path : str
        Path to the CSV file.
    chunksize : int
        Number of rows per chunk.
    **kwargs : dict
        Additional arguments for `pandas.read_csv`.

    Yields
    ------
    ReadCsvResult
        A named tuple containing bad lines and the corresponding DataFrame chunk.

    Example
    -------
    >>> import abraxos

    >>> for bad_lines, df in abraxos.read_csv_chunks('bad.csv', 4):
    ...     print(bad_lines)
    ...     print(df)
    [['', '', '', 'd', '', 'f', '', '', '', '', 'f', '', '', '', '']]
        id  name    age
    0   1   Joe     38
    1   2   James   31
    2   3   Jordon  2
    [['', 'f', 'f', '5', '6', '7', '8']]
        id  name    age
    3   2   Jasper  31
    4   1   Jill    38
    """
    bad_lines: list[list[str]] = []
    kwargs.update({"on_bad_lines": bad_lines.append, "engine": "python"})
    
    chunks = pd.read_csv(path, chunksize=chunksize, **kwargs)
    for chunk in chunks:
        yield ReadCsvResult(bad_lines.copy(), chunk)
        bad_lines.clear()


def read_csv_all(
    path: str,
    **kwargs
) -> ReadCsvResult:
    """
    Reads a CSV file and captures bad lines separately.

    Parameters
    ----------
    path : str
        Path to the CSV file.
    **kwargs : dict
        Additional arguments for `pandas.read_csv`.

    Returns
    -------
    ReadCsvResult

    Example
    -------
    >>> import abraxos

    >>> bad_lines, good_df = abraxos.read_csv('people.csv')
    >>> bad_lines
    [['', '', '', 'd', '', 'f', '', '', '', '', 'f', '', '', '', ''],
    ['', 'f', 'f', '5', '6', '7', '8']]
    >>> good_df
        id   name    age
    0   1    Joe     38
    1   2    James   31
    2   3    Jordon  2
    3   2    Jasper  31
    4   1    Jill    38
    """  
    bad_lines: list[list[str]] = []
    kwargs.update({"on_bad_lines": bad_lines.append, "engine": "python"})
    df: pd.DataFrame = pd.read_csv(path, **kwargs)
    return ReadCsvResult(bad_lines, df)


def read_csv(
    path: str,
    *,
    chunksize: int | None = None,
    **kwargs
) -> ReadCsvResult | a.Generator[ReadCsvResult, None, None]:
    """
    Reads a CSV file and captures bad lines separately.

    Parameters
    ----------
    path : str
        Path to the CSV file.
    chunksize : int, optional
        Number of rows per chunk. If None, reads the entire file at once.
    **kwargs : dict
        Additional arguments for `pandas.read_csv`.

    Returns
    -------
    ReadCsvResult or Generator
        If `chunksize` is specified, returns a generator yielding `ReadCsvResult`,
        otherwise returns a single `ReadCsvResult`.
    Example
    -------
    >>> import abraxos

    >>> bad_lines, good_df = abraxos.read_csv('people.csv')
    >>> bad_lines
    [['', '', '', 'd', '', 'f', '', '', '', '', 'f', '', '', '', ''],
    ['', 'f', 'f', '5', '6', '7', '8']]
    >>> good_df
        id   name    age
    0   1    Joe     38
    1   2    James   31
    2   3    Jordon  2
    3   2    Jasper  31
    4   1    Jill    38
    """
    if chunksize is not None:
        return read_csv_chunks(path, chunksize, **kwargs)
    
    return read_csv_all(path, **kwargs)


import collections.abc as a
import typing as t

import pandas as pd

from abraxos import split
 

class TransformResult(t.NamedTuple):
    errors: list[Exception]
    errored_df: pd.DataFrame
    success_df: pd.DataFrame


def transform(
    df: pd.DataFrame,
    transformer: a.Callable[[pd.DataFrame], pd.DataFrame]
) -> TransformResult:
    errors: list[Exception] = []
    errored_dfs: list[pd.DataFrame] = []
    success_dfs: list[pd.DataFrame] = []
    try:
        return TransformResult([], df[0:0], transformer(df))
    except Exception as e:
        if len(df) > 1:
            df1, df2 =  split.split_df(df)
            errors1, e_df1, s_df1 = transform(df1, transformer)
            errors2, e_df2, s_df2 = transform(df2, transformer)
            errors.extend(errors1 + errors2)
            errored_dfs.extend([e_df1, e_df2])
            success_dfs.extend([s_df1, s_df2])
        else:
            try:
                return TransformResult([], df[0:0], transformer(df))
            except Exception as e:
                return TransformResult([e], df, df[0:0])
            
    errored_df: pd.DataFrame = pd.concat(errored_dfs)
    success_df: pd.DataFrame = pd.concat(success_dfs)
    return TransformResult(errors, errored_df, success_df)


import typing as t
import collections.abc as a

import pandas as pd
import sqlalchemy as sa
import numpy as np

from abraxos import split
from abraxos import fill


class ToSqlResult(t.NamedTuple):
    errors: list
    errored_df: pd.DataFrame
    success_df: pd.DataFrame
 

def to_sql(
    df: pd.DataFrame,
    name: str,
    con: sa.engine.Connection | sa.engine.Engine,
    *,
    if_exists: t.Literal['fail', 'replace', 'append'] = 'append',
    index: bool = False,
    **kwargs
) -> ToSqlResult:
    errors: list[Exception] = []
    errored_dfs: list[pd.DataFrame] = [df[0:0], ]
    success_dfs: list[pd.DataFrame] = [df[0:0], ]
    try:
        df.to_sql(name, con, if_exists=if_exists, index=index, method='multi', **kwargs)
        return ToSqlResult([], df[0:0], df)
    except Exception as e:
        if len(df) > 1:
            df1, df2 = split.split_df(df)
            errors1, errored_df1, success_df1 = to_sql(df1, name, con, if_exists=if_exists, index=index, **kwargs)
            errors2, errored_df2, success_df2 = to_sql(df2, name, con, if_exists=if_exists, index=index, **kwargs)
            errors.extend(errors1 + errors2)
            errored_dfs.extend([errored_df1, errored_df2])
            success_dfs.extend([success_df1, success_df2])
        else:
            try:
                df.to_sql(name, con, if_exists=if_exists, index=index, method='multi', **kwargs)
                return ToSqlResult([], df[0:0], df)
            except Exception as e:
                return ToSqlResult([e], df, df[0:0])

    return ToSqlResult(errors, pd.concat(errored_dfs), pd.concat(success_dfs))


def to_records(df: pd.DataFrame) -> list[dict]:
    df = df.fillna(np.nan).replace([np.nan], [None])
    return df.to_dict('records')


def insert_df(
    df: pd.DataFrame,
    connection: sa.engine.Connection,
    sql_query: sa.Insert
) -> ToSqlResult:
    records: list[dict] = to_records(df)
    connection.execute(sql_query, records)
    return ToSqlResult([], df[0:0], df)


def use_sql(
    df: pd.DataFrame,
    connection: sa.engine.Connection,
    sql_query: sa.Insert
) -> ToSqlResult:
    """
    User user provided SQL Insert to inser DataFrame records.
    """
    errors: list[Exception] = []
    errored_dfs: list[pd.DataFrame] = [df[0:0], ]
    success_dfs: list[pd.DataFrame] = [df[0:0], ]
    try:
        return insert_df(df, connection, sql_query)
    except Exception as e:
        if len(df) > 1:
            df1, df2 = split.split_df(df)
            errors1, errored_df1, success_df1 = use_sql(df1, connection, sql_query)
            errors2, errored_df2, success_df2 = use_sql(df2, connection, sql_query)
            errors.extend(errors1 + errors2)
            errored_dfs.extend([errored_df1, errored_df2])
            success_dfs.extend([success_df1, success_df2])
        else:
            try:
                return insert_df(df, connection, sql_query)
            except Exception as e:
                return ToSqlResult([e], df, df[0:0])

    return ToSqlResult(errors, pd.concat(errored_dfs), pd.concat(success_dfs))